#define vec2 float2
#define vec3 float3
#define vec4 float4
#define mix lerp

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------


texture  f_texture;
sampler3D f_text =
sampler_state
{
    Texture = <f_texture>;
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    
    AddressU = WRAP;
    AddressV = WRAP;
	
};


#define my_color vec3(0.9,0.9,0.8)

#define z_near 0.3
#define z_far 60.0


float level = 0.0;
float step_length = 0.005;
vec3 box1 = { -0.990f, -0.990f, -0.990f};
vec3 box2 = { 0.990f, 0.990f, 0.990f};

float3 g_LightDir = { 1.0f, 0.0f, 0.0f};                  // Light's direction in world space
float4   g_vEye = { -5.0f, 0.0f, 0.0f,0.0f};                    // Camera's location

// Matrices:
float4x4 g_mWorld;                  // World matrix for object
float4x4 g_mWorldViewProjection;    // World * View * Projection matrix


float Equ(in float3 arg)
{
	return tex3Dlod(f_text, vec4(arg*0.5+vec3(0.5,0.5,0.5),0.0)).w;

/*	
	
	arg*=5.0f;
	float x = arg.x, y = arg.y, z = arg.z, T = 1.6180339887;

	return  -2.4 + cos ( x + T * y ) - cos ( x - T * y ) - cos ( y + T * z ) - cos ( y - T * z ) - cos ( z - T * x ) - cos ( z + T * x );
	*/
				
}



const float delta=1.0/256.0;

float3 GradEqu(in float3 arg)
{
		float3 res;
		res.x = Equ(float3(arg.x+delta,arg.y,arg.z))-Equ(float3(arg.x-delta,arg.y,arg.z));
		res.y = Equ(float3(arg.x,arg.y+delta,arg.z))-Equ(float3(arg.x,arg.y-delta,arg.z));
		res.z = Equ(float3(arg.x,arg.y,arg.z+delta))-Equ(float3(arg.x,arg.y,arg.z-delta));
		return normalize(res);
}


#define K_A 0.1
#define K_D 0.9
#define K_S 0.3
#define PPPP 16.0

float3 Phong ( float3 point1, float3 normal1, float3 color1)
{
	float diffuse = dot ( g_LightDir, normal1 )*0.5+0.5;
	float3 reflect1 = reflect ( normalize(point1-g_vEye.xyz), normal1 );
	float specular = pow ( abs(max (0.0f, dot ( reflect1, g_LightDir ) )), PPPP );
	return K_A * color1 + diffuse * ( K_D * color1 + K_S * specular * float3(1.0,1.0,1.0) );
}

float GetDepth(vec3 ps,vec3 pos, vec3 nav)
{
	ps -= pos;
	return z_far/(z_far-z_near) - (z_far*z_near/(z_far-z_near))/(dot(ps,nav));
}



bool IntersectBox ( float3 Origin, float3 Direction, float3 minimum, float3 maximum, out float start, out float final )
{
	float3 OMAX = ( minimum - Origin ) / Direction;
	float3 OMIN = ( maximum - Origin ) / Direction;
	float3 MAX = max ( OMAX, OMIN );
	float3 MIN = min ( OMAX, OMIN );
	final = min ( MAX.x, min ( MAX.y, MAX.z ) );
	start = max ( max ( MIN.x, 0.0), max ( MIN.y, MIN.z ) );	
	return final > start;
}

//-----------------------------------------------------------------------------
// Vertex shader output structure
//-----------------------------------------------------------------------------
struct VS_OUTPUT
{
    float4 position          : POSITION;
	float4 vPositionWS       : TEXCOORD0;
};


//-----------------------------------------------------------------------------
// Name: RenderSceneVS
// Type: Vertex shader                                      
// Desc: This shader computes standard transform and lighting
//-----------------------------------------------------------------------------
VS_OUTPUT RenderSceneVS( float4 vPos : POSITION )
{
    VS_OUTPUT Output;
  
    Output.position = mul(vPos, g_mWorldViewProjection);
    Output.vPositionWS = mul( vPos, g_mWorld);
    
    return Output;    
}


//-----------------------------------------------------------------------------
// Pixel shader output structure
//-----------------------------------------------------------------------------

struct PS_OUTPUT
{
    float4 RGBColor : COLOR0;  // Pixel color    
	float depth : DEPTH;
};

struct PS_INPUT
{
	float4 vPositionWS       : TEXCOORD0;
};


PS_OUTPUT RenderScenePS( VS_OUTPUT In ) 
{ 
    PS_OUTPUT Out;

	Out.RGBColor = float4( 0.0, 0.0, 0.0, 0.0 );
	Out.depth = 1.0;
   
	vec4 ps = vec4(g_vEye.xyz,0.0);
	float start,final;
	vec4 nav=vec4(normalize(In.vPositionWS.xyz-g_vEye.xyz),1.0);
	vec4 step = nav*step_length;
	float e,e0;

	if(IntersectBox (g_vEye.xyz, nav.xyz, box1, box2,start,final))
	{
		ps += nav*start;
		
		e = Equ(ps.xyz);
		ps+=step;

		if(e < level)
		{
			for(int i=0;i<3;i++)
			while(ps.w<=final)
			{
				e = Equ(ps.xyz);
				if(e >= level)	break;
				ps+=step;
			}
			if(ps.w>final){ps = vec4(g_vEye.xyz,0.0)+nav*final;e = Equ(ps.xyz);}
			
			if(e >= level)
			{
				e0 = Equ(ps.xyz-step.xyz);
				ps += step*((level-e)/(e-e0));
				vec3 norm = -GradEqu(ps.xyz);
				Out.RGBColor.xyz = Phong(ps.xyz,norm,my_color);
				Out.RGBColor.a = 1.0;

				Out.depth = GetDepth(ps.xyz,g_vEye.xyz,nav.xyz);
			}
		}else
		{
			Out.depth = GetDepth(ps.xyz,g_vEye.xyz,nav.xyz);
			ps -= step*1.1;
			vec3 norm = normalize(ps.xyz-clamp(ps.xyz,box1,box2));
			Out.RGBColor.xyz = Phong(ps.xyz,norm,my_color);
			Out.RGBColor.a = 1.0;
			
			
		}
		
			

	}
			
	

	return Out;
}


technique RenderScene
{
    pass P0
    {          
        VertexShader = compile vs_3_0 RenderSceneVS();
        PixelShader  = compile ps_3_0 RenderScenePS();
    }
}
